/*

======================================================================

  Seyon is Copyright (c) 1992 of Muhammad M. Saggaf. Seyon is not
  public domain. Permission is granted to use and distribute Seyon
  freely for any use and to sell it at any price without reference to
  the copyright owner provided that in all above cases Seyon is intact
  and is not made part of any program either in whole or in part and
  that this copyright notice is included with Seyon. Permission is
  also granted to modify the source as long as the modified source is
  not distributed.

======================================================================

*/

#include <stdio.h>
#include <signal.h>
#include <fcntl.h>

#include <string.h>
#include <sys/types.h>
#include <errno.h>

#include <X11/Intrinsic.h>

#include "seyon.h"
#include "SeDecl.h"
#include "SeDeclX.h"

#ifdef HAVE_TERMIOS
#include <termios.h>
#else
#ifdef HAVE_TERMIO
#include <termio.h>
#endif
#endif

#ifndef O_NDELAY
#ifdef O_NONBLOCK
#define O_NDELAY O_NONBLOCK
#endif
#endif

/*
    The dial() routine uses these two defines.

    DIALSTR is a sprintf format string that assumes a HAYES-compatible
    modem.

    HAYSATT is the HAYES "attention" signal (used if DTR_DROPS_CARRIER
    is not set to 1).

    HAYSHUP is the HAYES "hangup" command (used if DTR_DROPS_CARRIER is not
    set to 1).

    MDELAY is the delay in the output because (on my modem) the command
    would be ignored if sent at full speed.  Change for other setups.
    (This setting is for U.S. Robotics Password Modem).
*/

#define HAYSATT "+++"		/* Hayes "attention" signal */
#define	HAYSHUP	"ATH\r"		/* Hayes "hang up" command */
#define MDELAY  20000           /* delay for output to modem itself */

/* globals */

static int mfd = -1;            /* modem port file descriptor */

#ifdef HAVE_TERMIOS
static struct termios pmode;	/* modem device control string */
#else
#ifdef HAVE_TERMIO
static struct termio pmode;		/* modem device control string */
#endif
#endif

static char port[NMSIZE];       /* modem port device file string */
static int baudrate = B9600;	/* baud rate */

flush_modem()
{
  io_flush(mfd);
}

void send_break()
{
  io_send_break(mfd);
  SeyonMessage("SENT BREAK");
}

get_modem_attr()
{
  io_get_attr(mfd, &pmode);
}

sync_modem_attr()
{
  get_modem_attr();
  baudrate = io_get_speed(&pmode);
}

int get_modem_fio()
{
  return fcntl(mfd, F_GETFL);
}

set_modem_fio(fio)
	 int fio;
{
  fcntl(mfd, F_SETFL, fio);
}

char *read_line()
{
  static char buf[WBSIZE];
  int c, i;

  do {
	c = readbyte(0);
  } while(c <= 0 || c == '\r' || c == '\n');

  for(i = 0; c != '\r' && c != '\n'; i++) {
	buf[i] = c;
	while((c = readbyte(0)) <= 0);
  }
  i++;
  buf[i] = '\0';
  
  return buf;
}

void cancel_dial(verbose)
	 short verbose;
{
  if (mfd == -1)
	return;

  if(verbose)
	SeyonMessage("CANCELED");

  sendbyte(qres.dialCancel);
  flush_modem();
}

int cpmode()
{
    pmode.c_iflag |= IGNBRK;
    pmode.c_lflag = 0;

#ifdef XCLUDE
    pmode.c_lflag |= XCLUDE;
#endif

    pmode.c_oflag &= ~OPOST;	/* transparent output */
    pmode.c_cflag = baudrate | CS8 | CREAD | CLOCAL;
    pmode.c_cc[VMIN] = 1;		/* this many chars satisfies reads */
    pmode.c_cc[VTIME] = 1;		/* or in this many tenths of seconds */

    xc_setflow();
    set_rtscts();
}

set_rtscts()
{
#ifdef CRTSCTS
  if (qres.rtsctsFlowControl) {
	pmode.c_cflag |= CRTSCTS;

#ifdef CNORTSCTS
	pmode.c_cflag &= ~CNORTSCTS;
#endif

  } else {
	pmode.c_cflag &= ~CRTSCTS;

#ifdef CNORTSCTS
	pmode.c_cflag |= CNORTSCTS;
#endif

  }
#endif
  
  if (mfd != -1)
	io_set_attr(mfd, &pmode);
}

xc_setflow()
{
  if (qres.xonxoffFlowControl) {

#if defined(HAVE_TERMIOS) || defined(HAVE_TERMIO)
	pmode.c_iflag |= IXON | IXOFF;

#ifdef HAVE_TERMIO
	pmode.c_iflag &= ~IXANY;
#endif
#endif

  } else

#if defined(HAVE_TERMIOS) || defined(HAVE_TERMIO)
	pmode.c_iflag &= ~(IXON | IXOFF);

#ifdef HAVE_TERMIO
	pmode.c_iflag &= ~IXANY;
#endif
#endif
  
  if (mfd != -1)
	io_set_attr(mfd, &pmode);
}

char *mport(s)  /* get/set port string */
char *s;
{
    if(s != NULL && mfd == -1)
        strcpy(port, s);
    return(port);
}

/* Get/set the baud rate of the modem port. If the port hasn't been opened yet,
 * just store in pmode for mopen() to use when it opens the port.
 */
int mbaud(s)
char *s;
{
  cpmode();

  if(s != NULL){
								/* this gives a more limited, realistic */
	switch (atoi(s)) {			/* range than in sgtty.h */
	case   300: baudrate = B300;   break;
	case  1200: baudrate = B1200;  break;
	case  2400: baudrate = B2400;  break;
	case  4800: baudrate = B4800;  break;
	case  9600: baudrate = B9600;  break;
	case 19200: baudrate = B19200; break;
	case 38400: baudrate = B38400; break;
	default: return(-1);
	}
	io_set_speed(&pmode, baudrate);

	if(mfd != -1)
	  io_set_attr(mfd, &pmode);	  
  }
  
  switch(io_get_speed(&pmode)) {
  case   B300: return(300);
  case  B1200: return(1200);
  case  B2400: return(2400);
  case  B4800: return(4800);
  case  B9600: return(9600);
  case B19200: return(19200);
  case B38400: return(38400);
  }

  se_error("impossible error in baud rate");
  return(0);
}

/*
    The following routine is used to hang up the modem.  This is
    accomplished by setting the baud rate to 0.  According to my
    documentation on termio, setting the baud rate to zero will
    result in DTR not being asserted.  This hangs up some (most?)
    modems.  If not, the second part of the routine sends the Hayes
    modem "escape" and then a hangup command.
*/

void hangup()
{
    if (mfd == -1)
	return;

    SeyonMessage("HANGUP");
    fflush(stdout);

#ifdef DTR_DROPS_CARRIER
	io_set_speed(&pmode, B0); /* set baud 0 (drop DTR) */
	io_set_attr(mfd, &pmode);

    sleep(1);			/* wait a minute */

	io_set_speed(&pmode, baudrate); /* reset baud rate */
	io_set_attr(mfd, &pmode);

#else	/* use Hayes command */
    sleep(2);			/* Allow for "escape guard time" */
    send_slowly(HAYSATT);	/* Send modem escape command */

    sleep(3);			/* More "escape guard time" */
    send_slowly(HAYSHUP);	/* Send hangup command */
#endif
}

/* Opens the modem port and configures it. If the port string is
 * already defined it will use that as the modem port;  otherwise it
 * gets the environment variable MODEM. Returns 0 for success
 * or -1 on error.
 */

static int	  opened = 0;

int open_modem()
{
  char *p, *getenv();
  int oldflags;
  
  cpmode();
  
  if(port[0] == '\0'){
	if((p = getenv("MODEM")) == NULL) {
	  se_error("no modem device is specified");
	  return(-1);
	}
	strcpy(port, p);
  }
  
  if (lock_tty()) {
	show(">> Error: modem device is locked.");
	return -1;
  }
  
  /* Need O_NDELAY to get the file open before we have carrier */
  if((mfd = open(port, O_RDWR | O_NDELAY)) < 0) {
	se_error("could not open modem device");
	se_perror(port);
	unlock_tty();
	return -1;
  }
  
  /* Now, we must reset the O_NDELAY mode so that read() works correctly */
  if ( ((oldflags = fcntl(mfd, F_GETFL, 0)) == -1) ||
	  (fcntl(mfd, F_SETFL, oldflags & ~O_NDELAY) == -1) ) {
	se_perror("open_modem could not reset O_NDELAY");
	unlock_tty();
	return -1;
  }
  
  io_flush(mfd);
  io_set_attr(mfd, &pmode);

  if (mbaud(qres.defaultBPS) == -1)
	se_warningf("invalid default BPS value: %s", qres.defaultBPS);
  
  opened++;
  
  return(0);
}

close_modem()
{
  close(mfd);
}

#ifndef HAVE_DUP2		/* For those that do not have dup2()... */
int dup2(oldfd, newfd)
int oldfd, newfd;
{
        if (fcntl(oldfd, F_GETFL, 0) == -1)     /* Valid file descriptor? */
                return (-1);                    /* No, return an error. */
        close(newfd);                           /* Ensure newfd is closed */
        return (fcntl(oldfd, F_DUPFD, newfd));  /* Dup oldfd into newfd */
}
#endif /* HAVE_DUP2  Thanks to Bill Allie CIS: 76703,2061 */

/* Attach standard input and output to the modem port. This only gets called
 * after a fork by the child process; which then exec's a program that uses
 * standard i/o for some data transfer protocol. (To put this here is actually
 * a kludge, but I wanted to keep the modem-specific stuff in a black box.)
 */
mattach()       /* attach standard i/o to port */
{
    dup2(mfd, 0);       /* close local stdin and connect to port */
    dup2(mfd, 1);       /* close local stdout and connect to port */

    close(mfd);         /* close the old port descriptor */
}

int readbyte(seconds)
int seconds;
{
    return trminp(mfd, seconds);
}

/*
    Read a byte using bitmask
*/

read_mbyte(secs)
{
    int bitmask, c;

	bitmask = (qres.sevenBitMask) ? SBIT : EBIT;

    return (c = readbyte(secs)) == -1 ? -1 : c & bitmask;
}

/* Output a byte to the modem port.
 * All data sent to the modem is output through this routine.
 */
sendbyte(ch)
int ch;
{
    char c = ch & EBIT;

    write(mfd, &c, 1);
}

send_mbyte(ch)
int ch;
{
    int bitmask;

	bitmask = (qres.sevenBitMask) ? SBIT : EBIT;

    sendbyte(ch & bitmask);
}

/* Dial a phone number, using proper format and delay.
 */

static char *last_nbr = NULL;
char *strdup();

dial(s)
char *s;
{
    char buffer[WBSIZE];

    if (last_nbr) 
	free(last_nbr);

    last_nbr = strdup(s);

	sendf_slowly("\r%s %s%s", qres.dialPrefix, s, qres.dialSuffix);
  }

redial(last_nbr)
 char *last_nbr;
{
    char *s;

    if (last_nbr == NULL) {
	fprintf(tfp,"REDIAL FAILURE\r\n");
	return 1;
    }

    s = strdup(last_nbr);
    dial(s);
    free(s);
    return 0;
}

sendf_slowly(format, a, b, c, d)
	 char *format, *a, *b, *c, *d;
{
  char buffer[SM_BUF];
  
  sprintf(buffer, format, a, b, c, d);
  send_slowly(buffer);
}

send_slowly(s)
 char *s;
{
  int i;

/*
  This busy-waiting, normally a bad idea on a multi-tasking system,
  was used because sleep(1) is way too much of a delay.
*/
  while(*s){
	send_mbyte(*s++);

	/* avoid busy waiting by using usleep if available.
	   adjust MDELAY if you're having trouble */

#ifdef HAVE_USLEEP
	usleep(MDELAY);
#else
	for(i = 0; i < MDELAY; i++);
#endif

  }
}

/*
    lock_tty() returns non-zero if the lock file exists (prevents Seyon from
    running).

    unlock_tty() deletes the lock file.

    Simple, eh?
*/

char lckf[40] = LOCKFILE;
char ltmp[40] = LTMPFILE;

lock_tty()
{
  int pid, lckpid, lfd;
  char pidstr[20];
  char lckpidstr[20];
  char *modemname;
  extern int errno;
  
  /*
   * Get our PID, and initialize the filename strings.
   */
  pid = getpid();
  sprintf(pidstr, "%d", pid);
  modemname = strrchr(port, '/');
  strcat(lckf, (modemname ? (modemname+1) : port));
  strcat(ltmp, pidstr);
  
  /*
   * Create the LTMP.<pid> file and scribble our PID in it.
   */
  unlink(ltmp);
  if ((lfd = creat(ltmp, 0644)) == -1) {
	showf("Can't creat(2) LTMP file %s", LTMPFILE);
	return -1;
  }
  sprintf(pidstr, "%10d\n", pid);
  write(lfd, pidstr, 11);
  close(lfd);
  
  /*
   * Attempt to link directly - if it works, we're done.
   */
 relink:
  if (link(ltmp, lckf) == 0) {
	unlink(ltmp);
	return 0;
  }
  
  /*
   * Oh, bother, there's a LCK..* file already; we must
   * now expend effort to learn if it's stale or not.
   */
  if ((lfd = open(lckf, O_RDONLY)) != -1) {
	if (read(lfd, lckpidstr, 11) == 11) {
	  lckpid = atol(lckpidstr);
	  if (kill(lckpid, 0) == 0) {
		se_errorf("%s locked by process %d", port, lckpid);
		unlink(ltmp);
		return -1;
	  }
	}
  }
  
  /*
   * The LCK..* file was stale.  Remove and retry.
   */
  if (unlink(lckf)) {
	fprintf(stderr, "Can't unlink(2) stale LCK?\r\n");
	unlink(ltmp);
	return -1;
  }
  goto relink;
  /*NOTREACHED*/
}

unlock_tty()
{
  unlink(lckf);
}
